/* -*- mode: Java; c-basic-offset: 4; indent-tabs-mode: nil; -*-  //------100-columns-wide------>|*/
/* Copyright (c) 2006 Extreme! Lab, Indiana University. All rights reserved.
 * This software is open source. See the bottom of this file for the license.
 * $Id: LeadDeploymentConfig.java,v 1.5 2006/11/20 03:24:50 aslom Exp $ */
package xsul.lead;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import javax.xml.namespace.QName;
import org.xmlpull.v1.builder.XmlBuilderException;
import org.xmlpull.v1.builder.XmlDocument;
import org.xmlpull.v1.builder.XmlElement;
import org.xmlpull.v1.builder.XmlInfosetBuilder;
import org.xmlpull.v1.builder.XmlNamespace;
import xsul.MLogger;
import xsul.XmlConstants;
import xsul.XsulException;
import xsul.ws_addressing.WsaEndpointReference;

/**
 * Represents configuration of LEAD services (such as for production, deployment, experimental setup)
 * and allows to chain configurations for default values and easy overriding.
 * One place to keep list of services and allows overriding and cascading of configurations.
 */
public class LeadDeploymentConfig extends BaseXmlElementAdapter {
    private final static MLogger logger = MLogger.getLogger();
    private final static XmlInfosetBuilder builder = XmlConstants.BUILDER;
    
    public final static QName TYPE = new QName(
        "http://lead.extreme.indiana.edu/namespaces/2006/lead-deployment-config/", "config");
    public final static XmlNamespace NS = builder.newNamespace("ldc", TYPE.getNamespaceURI());
    
    public final static String AUDITING_WSDL_URL = "auditing-wsdl-url";
    public final static String DSC_URL = "dsc-url";
    public final static String GFAC_URL = "gfac-url";
    public final static String GPEL_URL = "gpel-engine-url";
    public final static String MSG_BOX_URL = "msg-box-url";
    public final static String MSG_BROKER_URL = "msg-broker-url";
    public final static String MYLEAD_AGENT_URL = "mylead-agent-url";
    public final static String RESOURCE_CATALOG_URL = "resource-catalog-url";
    public final static String ERROR_SINK_EPR = "error-sink-epr";
    public final static String WCS_WSDL_URL = "wcs-wsdl-url";
    public final static String XBAYA_CODEBASE_URL = "xbaya-codebase-url";
        
    private LeadDeploymentConfig parentConfig;
    
    public LeadDeploymentConfig() {
        super(builder.newFragment(NS, TYPE.getLocalPart()));
    }
    
    public LeadDeploymentConfig(LeadDeploymentConfig parentConfig) {
        this();
        if(parentConfig == null) throw new IllegalArgumentException();
        this.parentConfig = parentConfig;
    }
    
    public LeadDeploymentConfig(XmlElement target) {
        super(target);
    }
    
    public LeadDeploymentConfig(XmlElement target, LeadDeploymentConfig parentConfig) {
        this(target);
        if(parentConfig == null) throw new IllegalArgumentException();
        this.parentConfig = parentConfig;
    }
    
    public void setParentConfig(LeadDeploymentConfig parentConfig) {
        this.parentConfig = parentConfig;
    }
    
    public LeadDeploymentConfig getParentConfig() {
        return parentConfig;
    }
    
    
    public void setAuditingWsdlUrl(URI url) { setUriValue(NS, AUDITING_WSDL_URL, url); }
    public URI getAuditingWsdlUrl() { return lookupUriValue(NS, AUDITING_WSDL_URL); }
    public void setBrokerUrl(URI url) { setUriValue(NS, MSG_BROKER_URL, url); }
    public URI getBrokerUrl() { return lookupUriValue(NS, MSG_BROKER_URL); }
    public void setDscUrl(URI url) { setUriValue(NS, DSC_URL, url); }
    public URI getDscUrl() { return lookupUriValue(NS, DSC_URL); }
    public void setGfacUrl(URI url) { setUriValue(NS, GFAC_URL, url); }
    public URI getGpelUrl() { return lookupUriValue(NS, GPEL_URL); }
    public void setGpelUrl(URI url) { setUriValue(NS, GPEL_URL, url); }
    public URI getGfacUrl() { return lookupUriValue(NS, GFAC_URL); }
    public void setMsgBoxUrl(URI url) { setUriValue(NS, MSG_BOX_URL, url); }
    public URI getMsgBoxUrl() { return lookupUriValue(NS, MSG_BOX_URL); }
    public void setMyLeadAgentUrl(URI url) { setUriValue(NS, MYLEAD_AGENT_URL, url); }
    public URI getMyLeadAgentUrl() { return lookupUriValue(NS, MYLEAD_AGENT_URL); }
    public void setResourceCatalogUrl(URI url) { setUriValue(NS, RESOURCE_CATALOG_URL, url); }
    public URI getResourceCatalogUrl() { return lookupUriValue(NS, RESOURCE_CATALOG_URL); }
    public void setWcsWsdlUrl(URI url) { setUriValue(NS, WCS_WSDL_URL, url); }
    public URI getWcsWsdlUrl() { return lookupUriValue(NS, WCS_WSDL_URL); }
    public void setXbayaCodebaseUrl(URI url) { setUriValue(NS, XBAYA_CODEBASE_URL, url); }
    public URI getXbayaCodebaseUrl() { return lookupUriValue(NS, XBAYA_CODEBASE_URL); }
    
    public void setErrorSink(WsaEndpointReference epr) {
        setEpr(epr, NS, ERROR_SINK_EPR);
    }
    
    public WsaEndpointReference getErrorSink() {
        return lookupEpr(NS, ERROR_SINK_EPR);
    }
    
    protected String lookupStringValue(XmlNamespace NS, String name) {
        String s = super.lookupStringValue(NS, name);
        if(s == null) {
            if(parentConfig != null) { // chain
                return parentConfig.lookupStringValue(NS, name);
            }
        }
        return s;
    }
    
    protected WsaEndpointReference lookupEpr(XmlNamespace ns, String name) {
        WsaEndpointReference epr = super.lookupEpr(ns, name);
        if(epr == null) {
            if(parentConfig != null) { // chain
                return parentConfig.lookupEpr(ns, name);
            }
        }
        return epr;
    }
    
    
    public static LeadDeploymentConfig loadConfig(URI base, URI location) {
        LeadDeploymentConfig config;
        URL url;
        try {
            url = base != null ? new URL(base.toURL(), location.toASCIIString()) : location.toURL();
            URLConnection conn = url.openConnection();
            XmlDocument doc = builder.parseInputStream(conn.getInputStream(),
                                                       conn.getContentEncoding());
            config = new LeadDeploymentConfig(doc.getDocumentElement());
        } catch(XmlBuilderException e) {
            throw new XsulException("could not load config from "+location, e);
        } catch (IOException e) {
            throw new XsulException("could not load config from "+location, e);
        }
        return config;
    }
    
    
    public static LeadDeploymentConfig loadConfig(Class classloader, String configlLoc) {
        URI base = ((new File(".")).toURI());
        URI wsdlLoclUri;
        try {
            wsdlLoclUri = new URI(configlLoc);
        } catch (URISyntaxException e) {
            throw new XsulException(
                "location of LEAD config file must be a correct URI, could not parse '"+configlLoc+"'", e);
        }
        LeadDeploymentConfig config;
        try {
            config = loadConfig(base, wsdlLoclUri);
        } catch(XsulException we) {
            Throwable  t = we.getCause();
            if(t instanceof FileNotFoundException) {
                // let try loading WSDL from CLASSPATH
                String locationOnClassapth = configlLoc;
                try {
                    if(configlLoc.charAt(0) != '/') {
                        locationOnClassapth = '/' + locationOnClassapth;
                    }
                    URL resourceUrl = classloader.getResource(locationOnClassapth);
                    //System.err.println("loading resourceUrl="+resourceUrl);
                    if(resourceUrl != null) {
                        URI uri = URI.create(resourceUrl.toString());
                        config = loadConfig(null, uri);
                        //System.err.println("loading "+locationOnClassapth+" from  resourceUrl="+resourceUrl);
                        //URLConnection conn = resourceUrl.openConnection();
                        //XmlDocument doc = builder.parseInputStream(conn.getInputStream(),
                        //                                           conn.getContentEncoding());
                        //config = new LeadDeploymentConfig(doc.getDocumentElement());
                    } else {
                        throw we;
                    }
                } catch (XsulException e) {
                    //e.printStackTrace();
                    throw new XsulException("could not load config from classpath "+locationOnClassapth, e);
                    //throw e;
//                } catch (IOException e) {
//                    throw new XsulException("could not load config from classpath "+locationOnClassapth, e);
                } catch(IllegalArgumentException e) {
                    throw we;
                }
            } else {
                throw we;
            }
        }
        return config;
    }
    
    
}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 *
 * Copyright (c) 2006 The Trustees of Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1) All redistributions of source code must retain the above
 *    copyright notice, the list of authors in the original source
 *    code, this list of conditions and the disclaimer listed in this
 *    license;
 *
 * 2) All redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the disclaimer
 *    listed in this license in the documentation and/or other
 *    materials provided with the distribution;
 *
 * 3) Any documentation included with all redistributions must include
 *    the following acknowledgement:
 *
 *      "This product includes software developed by the Indiana
 *      University Extreme! Lab.  For further information please visit
 *      http://www.extreme.indiana.edu/"
 *
 *    Alternatively, this acknowledgment may appear in the software
 *    itself, and wherever such third-party acknowledgments normally
 *    appear.
 *
 * 4) The name "Indiana University" or "Indiana University
 *    Extreme! Lab" shall not be used to endorse or promote
 *    products derived from this software without prior written
 *    permission from Indiana University.  For written permission,
 *    please contact http://www.extreme.indiana.edu/.
 *
 * 5) Products derived from this software may not use "Indiana
 *    University" name nor may "Indiana University" appear in their name,
 *    without prior written permission of the Indiana University.
 *
 * Indiana University provides no reassurances that the source code
 * provided does not infringe the patent or any other intellectual
 * property rights of any other entity.  Indiana University disclaims any
 * liability to any recipient for claims brought by any other entity
 * based on infringement of intellectual property rights or otherwise.
 *
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
 * NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
 * UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
 * SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
 * OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
 * SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
 * DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
 * RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
 * AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
 * SOFTWARE.
 */

